@@ -16,9 +16,12 @@ from collections import OrderedDict as ODict
 import pathlib
 import re
 import textwrap
+import typing
 from typing import List, NamedTuple, Optional, Union
 from . import message
 from . import util
+if typing.TYPE_CHECKING:
+    from .code_collections import CodeCollection
 
 
 
@@ -295,6 +298,7 @@ class Options(dict):
         code_chunk = self.code_chunk
         custom_options = self.custom_options
         if self['lang'] is None:
+            self['inherited_lang'] = True
             self['lang'] = code_chunk.copy_chunks[0].options['lang']
         if code_chunk.inline:
             if code_chunk.command == 'paste' and 'show' not in custom_options:
@@ -319,7 +323,7 @@ class Options(dict):
                             for kw in ('first_number', 'line_numbers', 'rewrap_lines', 'rewrap_width', 'expand_tabs', 'tab_size')])
     _first_chunk_execute_keywords = set(['executable', 'jupyter_kernel'])
     _first_chunk_save_keywords = set(['save', 'save_as'])
-    _first_chunk_other_keywords = set(['jupyter_timeout', 'live_output'])
+    _first_chunk_other_keywords = set(['executable_opts', 'args', 'jupyter_timeout', 'live_output'])
     _first_chunk_keywords = _first_chunk_execute_keywords | _first_chunk_save_keywords | _first_chunk_other_keywords
 
     keywords = _base_keywords | _layout_keywords | _first_chunk_keywords
@@ -332,6 +336,7 @@ class Options(dict):
     _default_inline_options = {'complete': True,
                                'example': False,
                                'lang': None,
+                               'inherited_lang': False,
                                'outside_main': False}
     _default_block_options = _default_inline_options.copy()
     _default_block_options.update({'code_first_number': 'next',
@@ -429,6 +434,8 @@ class Options(dict):
             self['first_chunk_options'][key] = value
 
     _option_executable = _option_first_chunk_string_error
+    _option_executable_opts = _option_first_chunk_string_error
+    _option_args = _option_first_chunk_string_error
     _option_jupyter_kernel = _option_first_chunk_string_error
     _option_jupyter_timeout = _option_first_chunk_int_error
     _option_save = _option_first_chunk_bool_error
@@ -772,11 +779,11 @@ class CodeChunk(object):
         self.inline = inline
 
         # Check for len(code_lines) > 1 for inline later
-        self._code = None
+        self._code_str = None
         if isinstance(code, list):
             code_lines = code
         else:
-            code_lines = util.splitlines_lf(code) or ['']
+            code_lines = util.splitlines_lf(code, last_lf_stripped=True) or ['']
         if 'copy' not in custom_options and 'include' not in custom_options:
             if inline and len(code_lines) > 1:
                 self.errors.append(message.SourceError('Inline code cannot be longer that 1 line'))
@@ -809,17 +816,14 @@ class CodeChunk(object):
         if command == 'paste':
             if 'copy' not in custom_options:
                 self.errors.append(message.SourceError('Command "paste" cannot be used without specifying a target via "copy"'))
-            self.has_output = False
+            self.needs_to_copy = True
         else:
-            self.has_output = True  # Whether need output from copying
+            self.needs_to_copy = False
         if 'copy' in self.options:
             self.copy_chunks = []
 
-        if self.execute:
-            self.session = None
-        else:
-            self.source = None
-        self.index = None
+        self.code_collection: Optional[CodeCollection] = None
+        self.index: Optional[int] = None
         self.output_index = None
         self.stdout_lines = []
         self.stderr_lines = []
@@ -831,6 +835,18 @@ class CodeChunk(object):
         self.stdout_start_line_number = 1
         self.stderr_start_line_number = 1
 
+    @property
+    def session(self):
+        if self.execute:
+            return self.code_collection
+        raise TypeError
+
+    @property
+    def source(self):
+        if not self.execute:
+            return self.code_collection
+        raise TypeError
+
 
     def __pre_init__(self):
         '''
@@ -855,15 +871,33 @@ class CodeChunk(object):
 
 
     @property
-    def code(self):
-        code = self._code
-        if code is not None:
-            return code
+    def code_str(self):
+        code_str = self._code_str
+        if code_str is not None:
+            return code_str
         code = '\n'.join(self.code_lines)
-        self._code = code
+        self._code_str = code
         return code
 
 
+    def finalize_line_numbers(self, code_start_line_number):
+        # Line numbers can't be precalculated since they are determined by
+        # how a session is assembled across potentially multiple sources
+        self.code_start_line_number = code_start_line_number
+
+
+    @property
+    def attr_hash(self):
+        raise NotImplementedError
+
+    @property
+    def code_hash(self):
+        raise NotImplementedError
+
+    def only_code_output(self, format):
+        raise NotImplementedError
+
+
     def finalize_after_copy(self):
         '''
         Finalize options.  This can be redefined by subclasses so that they
@@ -912,8 +946,7 @@ class CodeChunk(object):
         if self.command == 'paste':
             if all(cc.command == 'code' for cc in copy_chunks):
                 # When possible, simplify the copying resolution process
-                self.has_output = True
-        self.code_start_line_number = copy_chunks[0].code_start_line_number
+                self.needs_to_copy = False
 
 
     def copy_output(self):
@@ -969,9 +1002,17 @@ class CodeChunk(object):
             self.expr_lines = copy_chunks[0].expr_lines
         self.stdout_start_line_number = copy_chunks[0].stdout_start_line_number
         self.stderr_start_line_number = copy_chunks[0].stderr_start_line_number
-        self.has_output = True
+        self.needs_to_copy = False
 
 
+    @property
+    def as_markup_lines(self):
+        raise NotImplementedError
+
+    @property
+    def as_example_markup_lines(self):
+        raise NotImplementedError
+
     def layout_output(self, output_type, output_format, lines=None):
         '''
         Layout all forms of output, except for rich output that is not
