@@ -48,6 +48,7 @@ class CodeProcessor(object):
                  origin_paths_for_cache: Optional[List[pathlib.Path]],
                  code_defaults: Optional[Dict],
                  session_defaults: Optional[Dict],
+                 only_code_output: bool,
                  progress: Progress):
         self.code_chunks = code_chunks
         self.cross_origin_sessions = cross_origin_sessions
@@ -62,6 +63,7 @@ class CodeProcessor(object):
             self._origin_paths_for_cache_as_strings = [p.as_posix() for p in origin_paths_for_cache]
         self.code_defaults = code_defaults
         self.session_defaults = session_defaults
+        self._only_code_output = only_code_output
         self._progress = progress
 
         self._old_cache_index: Optional[Dict] = None
@@ -90,7 +92,7 @@ class CodeProcessor(object):
         if any(s.status.prevent_exec for s in self._sessions.values()):
             code ^= 0b00000100
         if (any(s.status.has_errors and not s.status.prevent_exec for s in self._sessions.values()) or
-                any(s.status.has_errors and not s.status.prevent_exec for s in self._sources.values())):
+                any(s.status.has_errors for s in self._sources.values())):
             code ^= 0b00001000
         if any(s.status.has_warnings for s in self._sessions.values()):
             code ^= 0b00010000
@@ -105,11 +107,13 @@ class CodeProcessor(object):
         impossible to determine until all code chunks are assembled.  Group
         code chunks into sessions and sources.
         '''
-        self._set_chunk_keys()
         self._index_named_code_chunks()
         self._resolve_code_copying()
-        self._create_sessions_and_sources()
         self._prep_cache()
+        self._create_sessions_and_sources()
+        if self._only_code_output and self._cached_sessions:
+            # Resolve output copying as early as possible
+            self._resolve_output_copying(from_cache=True)
 
 
     def exec(self):
@@ -119,13 +123,12 @@ class CodeProcessor(object):
         builtin_sessions = []
         jupyter_sessions = []
         for session in self._sessions.values():
-            if session.status.prevent_exec:
+            if not session.needs_exec:
                 continue
-            if not self._load_session_cache(session):
-                if session.jupyter_kernel is None:
-                    builtin_sessions.append(session)
-                else:
-                    jupyter_sessions.append(session)
+            if session.jupyter_kernel is None:
+                builtin_sessions.append(session)
+            else:
+                jupyter_sessions.append(session)
         # Use `atexit` to improve the odds of updating the cache index in the
         # event of an unexpected exit.  Any session caches that are
         # successfully updated can be used in the future, regardless of
@@ -178,24 +181,6 @@ class CodeProcessor(object):
         self._update_session_cache(session)
 
 
-    def _set_chunk_keys(self):
-        '''
-        Assign code chunk session/source keys.
-        '''
-        for cc in self.code_chunks:
-            if cc.execute:
-                if self.cross_origin_sessions:
-                    key = CodeKey(cc.options['lang'], cc.options['session'], 'session', None)
-                else:
-                    key = CodeKey(cc.options['lang'], cc.options['session'], 'session', cc.origin_name)
-            else:
-                if self.cross_origin_sessions:
-                    key = CodeKey(cc.options['lang'], cc.options['source'], 'source', None)
-                else:
-                    key = CodeKey(cc.options['lang'], cc.options['source'], 'source', cc.origin_name)
-            cc.key = key
-
-
     def _index_named_code_chunks(self):
         '''
         Index named code chunks, and mark code chunks with duplicate names.
@@ -293,7 +278,7 @@ class CodeProcessor(object):
             unresolved_chunks = still_unresolved_chunks
             still_unresolved_chunks = []
 
-    def _resolve_output_copying(self):
+    def _resolve_output_copying(self, *, from_cache: bool=False):
         '''
         For code chunks with copying, handle the output copying.  The output
         copying for commands like "paste" must be handled after code is
@@ -304,11 +289,11 @@ class CodeProcessor(object):
         # dependencies; that's already been handled in
         # `_resolve_code_copying()`.  Still need to check for errors that
         # prevent copying, since there can be runtime source errors or other
-        # errors related to output.  Code chunks with `.has_output == True`
-        # are copying code-only code chunks and thus have no output for
-        # copying.
+        # errors related to output.  Code chunks with
+        # `.needs_to_copy == False` are copying code-only code chunks and thus
+        # have no output for copying.
         unresolved_chunks = [cc for cc in self.code_chunks
-                             if cc.command == 'paste' and not cc.errors.prevent_exec and not cc.has_output]
+                             if cc.command == 'paste' and not cc.errors.prevent_exec and cc.needs_to_copy]
         still_unresolved_chunks = []
         while True:
             for cc in unresolved_chunks:
@@ -323,11 +308,17 @@ class CodeProcessor(object):
                     traceback = '\n'.join(traceback_list)
                     msg = f'Code chunk(s) have error(s) that prevent copying:\n{traceback}'
                     cc.errors.append(message.SourceError(msg))
-                elif any(not copied_cc.has_output for copied_cc in cc.copy_chunks):
+                    self._progress.source_chunk_complete(cc.source, chunk=cc)
+                    if all(not cc.needs_to_copy or cc.errors.prevent_exec for cc in cc.source.code_chunks):
+                        self._progress.source_finished(cc.source)
+                elif any(copied_cc.needs_to_copy for copied_cc in cc.copy_chunks):
                     still_unresolved_chunks.append(cc)
                 else:
                     cc.copy_output()
-            if not still_unresolved_chunks:
+                    self._progress.source_chunk_complete(cc.source, chunk=cc)
+                    if all(not cc.needs_to_copy or cc.errors.prevent_exec for cc in cc.source.code_chunks):
+                        self._progress.source_finished(cc.source)
+            if not still_unresolved_chunks or (from_cache and len(unresolved_chunks) == len(still_unresolved_chunks)):
                 break
             unresolved_chunks = still_unresolved_chunks
             still_unresolved_chunks = []
@@ -337,19 +328,56 @@ class CodeProcessor(object):
         '''
         Assemble code chunks into sessions and sources.
         '''
+        placeholder_lang_num = 0
         for cc in self.code_chunks:
+            if cc.execute:
+                code_collection_name = cc.options['session']
+                code_collection_type = 'session'
+            else:
+                code_collection_name = cc.options['source']
+                code_collection_type = 'source'
+            if self.cross_origin_sessions:
+                origin_name = None
+            else:
+                origin_name = cc.origin_name
+            cc.key = CodeKey(cc.options['lang'], code_collection_name, code_collection_type, origin_name)
+            if cc.options['inherited_lang']:
+                cc.options['placeholder_lang'] = f'{placeholder_lang_num}'
+                placeholder_lang_num += 1
             if cc.execute:
                 self._sessions[cc.key].append(cc)
             else:
                 self._sources[cc.key].append(cc)
         for session in self._sessions.values():
             session.finalize()
-            if not session.status.prevent_exec:
+            if session.status.prevent_exec:
+                session.needs_exec = False
+            else:
                 self._session_hash_root_sets[session.hash_root].add(session)
+                if self._load_session_cache(session):
+                    session.needs_exec = False
         for source in self._sources.values():
             source.finalize()
-        self._progress.register_sessions(self._sessions.values())
-        self._progress.register_sources(self._sources.values())
+
+        # Need to register collections before reporting any progress
+        self._progress.register_code_collections(sessions=self._sessions.values(), sources=self._sources.values())
+        for session in self._sessions.values():
+            if session.status.prevent_exec:
+                self._progress.session_errors_prevent_exec(session)
+                for cc in session.code_chunks:
+                    self._progress.session_chunk_complete_no_exec(session, chunk=cc)
+                self._progress.session_finished(session)
+            elif session in self._cached_sessions:
+                self._progress.session_load_cache(session)
+                for cc in session.code_chunks:
+                    self._progress.session_chunk_complete_no_exec(session, chunk=cc)
+                self._progress.session_finished(session)
+        for source in self._sources.values():
+            for cc in source.code_chunks:
+                if not cc.needs_to_copy or cc.errors.prevent_exec:
+                    self._progress.source_chunk_complete(source, chunk=cc)
+            if all(not cc.needs_to_copy or cc.errors.prevent_exec for cc in source.code_chunks):
+                self._progress.source_finished(source)
 
 
     def _prep_cache(self):
@@ -449,9 +477,9 @@ class CodeProcessor(object):
         of rich output that results in additional files, because that is done
         during cache prep when the cache index is loaded.
         '''
-        if self.no_cache:
+        if self.no_cache or session.status.prevent_exec:
             return False
-        if session.hash in self._cached_sessions:
+        if session in self._cached_sessions:
             return True
         session_cache_path = self._cache_key_path / f'{session.hash_root}.zip'
         try:
@@ -487,8 +515,6 @@ class CodeProcessor(object):
             for output_name, output in chunk_cache.items():
                 setattr(chunk, output_name, output)
         self._cached_sessions.add(session)
-        self._progress.session_load_cache(session)
-        self._progress.session_finished(session)
         return True
 
 
@@ -505,13 +531,15 @@ class CodeProcessor(object):
         output is preserved if at all possible in the event of an unexpected
         exit.
         '''
-        if self.no_cache:
+        if self.no_cache or update_session.status.prevent_caching:
             return
         hash_root_cache = {
             'codebraid_version': codebraid_version,
             'cache': {},
         }
         for session in self._session_hash_root_sets[update_session.hash_root]:
+            if session.status.prevent_caching:
+                continue
             session_code_chunks_cache = {}
             for index, chunk in enumerate(session.code_chunks):
                 chunk_cache = {}
